/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.io.Serializable;
import java.text.MessageFormat;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicLong;

import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalenderAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalenderAntwort;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTypLernParameter;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienException;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.ZustandsWechsel;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienSpeicherClient;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlLesenAnfrage;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Sammelt und kapselt die Ausgangsdaten für das Lernen von Ganglinien.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Ausgangsdaten {

	/**
	 * Sortiert absteigend nach der Priorität der Ereignistypen.
	 */
	private static class EreignisTypNachPrioritaet implements Comparator<EreignisTyp>, Serializable {

		private static final long serialVersionUID = 1L;

		@Override
		public int compare(final EreignisTyp t1, final EreignisTyp t2) {
			return Long.compare(t2.getPrioritaet(), t1.getPrioritaet());
		}

	}

	/** Der Logger der Klasse. */
	private static final Debug LOGGER = Debug.getLogger();

	/** Der Messquerschnitt für den gelernt wird. */
	private final MessQuerschnitt mq;

	/** Der gesamte Lernzeitraum, kann mehrere Tage umfassen. */
	private final Intervall lernzeitraum;

	/** Die historischen Ganglinien am Messquerschnitt. */
	private final List<MqGanglinie> historischeGanglinien = new ArrayList<>();

	private final AtomicLong ereigniskalenderAnfrageId = new AtomicLong(1);
	
	/**
	 * Die Ereignistypen im aktuellen Lernzeitraum, absteigend sortiert nach
	 * Priorität.
	 */
	private final SortedSet<EreignisTyp> ereignisTypen;

	private GlLernen applikation;

	private Map<ApproximationTyp, MqGanglinie> archivGanglinien = new LinkedHashMap<>();

	/**
	 * Initialisierung den Lernprozess.
	 *
	 * @param applikation  die Ganglinien-Lernen-Applikation
	 * @param mq           der Messquerschnitt für den die Ganglinien gelernt
	 *                     werden.
	 * @param lernzeitraum der Zeitraum in dem gelernt werden soll.
	 */
	public Ausgangsdaten(GlLernen applikation, final MessQuerschnitt mq, final Intervall lernzeitraum) {
		this.applikation = applikation;
		this.mq = mq;
		this.lernzeitraum = lernzeitraum;
		assert lernzeitraum.getLength() == TimeUnit.DAYS.toMillis(1) : "Der Lernzyklus muss genau einen Tag umfassen.";

		ereignisTypen = new TreeSet<>(new EreignisTypNachPrioritaet());
	}

	/**
	 * Ruft vom Datenverteiler die zum Lernen notwendigen Daten ab. Dazu gehören die
	 * Archivdaten und die Ereignisse im Lernzeitraum.
	 *
	 * @return {@code false}, wenn die verfügbaren Daten zum Lernen nicht
	 *         ausreichen.
	 */
	public boolean abrufen() {

		// Im Lernzeitraum anliegende Ereignisse abfragen
		ereignisseAbrufen();

		// Historische Ganglinien bestimmen
		try {
			GanglinienSpeicherClient ganglinienSpeicher = applikation.getGanglinienSpeicher();
			historischeGanglinien.addAll(ganglinienSpeicher.liesGanglinien(
					GlLesenAnfrage.of(ganglinienSpeicher.getAnfrager(), ganglinienSpeicher.getAnfrageId(), getMq()),
					Duration.ofMinutes(10)));
		} catch (TimeoutException | GanglinienException e) {
			LOGGER.warning(() -> e.getLocalizedMessage());
		}

		LOGGER.finest(
				() -> MessageFormat.format("{0}: Vorhandene historische Ganglinien: {1}", mq, historischeGanglinien));

		if (ereignisTypen.isEmpty()) {
			String message = MessageFormat.format("{0}: Es liegen keine Ereignisse im Lernzeitraum an.", mq);
			LOGGER.warning(() -> message);
			applikation.warnung(mq, message);
			return false;
		}

		final StringBuilder txt = new StringBuilder();
		for (EreignisTyp typ : ereignisTypen) {
			if (txt.length() > 0) {
				txt.append(", ");
			}
			txt.append(typ);
			txt.append(" (Priorität ");
			txt.append(typ.getPrioritaet());
			txt.append(")");
		}
		assert txt.length() > 0 : "Es müssen Ereignisse anliegen.";

		LOGGER.fine(() -> MessageFormat.format("{0}: Es liegen folgende Ereignistypen im Lernzyklus an: {1}", mq,
				txt.toString()));

		return true;
	}

	/**
	 * Fragt den Ereigniskalender nach Ereignissen im Lernzyklus.
	 */
	private void ereignisseAbrufen() {

		EreignisKalenderAnfrage anfrage = new EreignisKalenderAnfrage(getAnfrageKennung());
		anfrage.setIntervall(lernzeitraum);
		anfrage.addRaeumlicheGueltigkeit(mq.getStrassenSegmentPid());

		EreignisKalenderAntwort kalenderAntwort;
		try {
			kalenderAntwort = applikation.getEreignisKalender().kalenderAnfragen(anfrage, Duration.ofMinutes(5));
			kalenderAnwortEmpfangen(kalenderAntwort);
		} catch (@SuppressWarnings("unused") TimeoutException e) {
			String message = MessageFormat
					.format("Für den MQ '{0}' konnten keine Ereigniskalender-Informationen abgerufen werden", mq);
			LOGGER.warning(() -> message);
			applikation.warnung(mq, message);
		}
	}

	private String getAnfrageKennung() {
		return "GlLernen_" + ereigniskalenderAnfrageId.incrementAndGet() + "_" + lernzeitraum;
	}

	private void kalenderAnwortEmpfangen(EreignisKalenderAntwort antwort) {

		for (final ZustandsWechsel zustandsWechsel : antwort.getZustandsWechsel()) {

			long zustandsWechselInMillis = LocalDateTimeKonverter.toMillis(zustandsWechsel.getZeitpunkt());
			if (((zustandsWechselInMillis <= lernzeitraum.getStart()) && !zustandsWechsel.isVerkehrlichGueltig())
					|| ((zustandsWechselInMillis >= lernzeitraum.getEnd()) && zustandsWechsel.isVerkehrlichGueltig())) {
				// Ein endendes Ereignis zu Beginn des Lernzeitraum und ein
				// beginnendes Ereignis am Ende des Lernzeitraum wird
				// ignoriert.
				continue;
			}

			final EreignisTyp typ = zustandsWechsel.getEreignis().getEreignisTyp();

			EreignisTypLernParameter lernParameter = typ.getLernParameter();

			if (typ.getPrioritaet() > 0 && !lernParameter.isNull()) {
				ereignisTypen.add(typ);
			} else if (typ.getPrioritaet() <= 0) {
				String message = MessageFormat
						.format("{0}: Priorität des Ereignistyps '{1}' ist kleiner oder gleich 0.", mq, typ);
				LOGGER.warning(() -> message);
				applikation.warnung(mq, message);
			} else if (lernParameter.isNull()) {
				String message = MessageFormat.format("{0}: Parameter {2} für {1} ist ungültig.", mq, typ,
						lernParameter);
				LOGGER.warning(() -> message);
				applikation.warnung(mq, message);
			}
		}

		LOGGER.fine(() -> {
			final StringBuilder txt = new StringBuilder();
			for (EreignisTyp typ : ereignisTypen) {

				if (txt.length() > 0) {
					txt.append(", ");
				}

				txt.append(typ);
				txt.append(" (Priorität ");
				txt.append(typ.getPrioritaet());
				txt.append(")");
			}
			return MessageFormat.format("{0}: Es liegen folgende Ereignistypen im Lernzyklus an: {1}", mq,
					txt.toString());
		});
	}

	/**
	 * Gibt den Messquerschnitt zurück, für den die Daten abgerufen werden.
	 *
	 * @return der Messquerschnitt.
	 */
	public MessQuerschnitt getMq() {
		return mq;
	}

	/**
	 * Gibt den Lernzetraum zurück, für den die Daten abgerufen werden.
	 *
	 * @return der Lernzeitraum.
	 */
	public Intervall getLernzeitraum() {
		return lernzeitraum;
	}

	/**
	 * Gibt die Archivdaten im Lernzeitraum als Ganglinie zurück.
	 *
	 * @param verfahren das Approximationsverfahren.<br>
	 *                  z.B.: {@link ApproximationTyp#BSPLINE},
	 *                  {@link ApproximationTyp#CUBICSPLINE},
	 *                  {@link ApproximationTyp#POLYLINE},
	 *                  {@link ApproximationTyp#UNBESTIMMT}
	 * @return die Archivganglinie fuer B-Spline.
	 */
	public MqGanglinie getArchivGanglinie(final ApproximationTyp verfahren) {

		MqGanglinie gl = archivGanglinien.get(verfahren);
		if (gl == null) {
			final MqGanglinie neueGl = applikation.getArchiv()
					.bestimmeArchivGanglinie(GanglinienAspekte.getAspect(verfahren), mq, lernzeitraum);
//            neueGl.setApproximationDaK(verfahren);
			if (neueGl.hatMesswerte()) {
				Calendar calCubicSpline = Calendar.getInstance();
				calCubicSpline.setTimeInMillis(neueGl.getStuetzstellen().get(0).getZeitstempel());
				calCubicSpline.set(Calendar.HOUR_OF_DAY, 0);
				calCubicSpline.set(Calendar.MINUTE, 0);
				calCubicSpline.set(Calendar.SECOND, 0);
				calCubicSpline.set(Calendar.MILLISECOND, 0);
				long offsetCubicSpline = -calCubicSpline.getTimeInMillis();
				GanglinienMQOperationen.verschiebe(neueGl, offsetCubicSpline);
			}

			gl = neueGl;
			LOGGER.finer(() -> MessageFormat.format("{0}: Normierte Archivganglinie: {1}", neueGl.getMessQuerschnitt(),
					neueGl));
			archivGanglinien.put(verfahren, gl);
		}
		return gl;
	}

	/**
	 * Gibt die historischen Ganglinien am zu lernenden Messquerschnit zurück.
	 *
	 * @return die historischen Ganglinien.
	 */
	public List<MqGanglinie> getHistorischeGanglinien() {
		return historischeGanglinien;
	}

	/**
	 * Gibt die Ereignistypen im Lernzeitraum absteigend sortiert nach Priorität
	 * zurück.
	 *
	 * @return die anliegenden Ereignistypen.
	 */
	public SortedSet<EreignisTyp> getEreignisTypen() {
		return ereignisTypen;
	}

}
