/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.LinkedHashMap;
import java.util.Map;

import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

/**
 * Stellt für jedes Approximationsverfahren den Lern-Aspekt zur Verfügung.
 *
 * @author BitCtrl Systems GmbH, Thomas Thierfelder
 */
public final class GanglinienAspekte {

    private GanglinienAspekte() {
        // Es gibt keine Instanzen der Klasse
    }
    
    /**
     * Der Standardaspekt für alle Approximationsverfahren
     * <code>asp.agregation5Minuten</code>.
     **/
    private static final String ASP_AGREGATION5_MINUTEN = "asp.agregation5Minuten";

    /**
     * Kommandozeilenargument für den Aspekt der B-Spline-Approximation.
     */
    private static final String ARG_BSPLINE = "-bSpline=" + ASP_AGREGATION5_MINUTEN;

    /**
     * Kommandozeilenargument für den Aspekt der Cubic-Spline-Approximation.
     */
    private static final String ARG_CUBIC_SPLINE = "-cSpline=" + ASP_AGREGATION5_MINUTEN;

    /**
     * Kommandozeilenargument für den Aspekt der Polyline-Approximation.
     */
    private static final String ARG_POLYLINE = "-polyline=" + ASP_AGREGATION5_MINUTEN;

    /** Mappt das Approximationsverfahren auf die PID des Aspekts. */
    private static final Map<ApproximationTyp, String> ASPECTPID_MAP = new LinkedHashMap<>();

    /**
     * Initisiert dieses statische Objekt mit den Parametern aus der Kommandozeile.
     *
     * @param argumentList die Datenverteiler-Argumentliste, wie sie in
     *                     {@link StandardApplication#parseArguments(ArgumentList)}
     *                     uebergeben wird.
     */
    public static void initialisiere(final ArgumentList argumentList) {
        final String aspBSpline = argumentList.fetchArgument(ARG_BSPLINE).asString();
        final String aspCubicSpline = argumentList.fetchArgument(ARG_CUBIC_SPLINE).asString();
        final String aspPolyline = argumentList.fetchArgument(ARG_POLYLINE).asString();

        ASPECTPID_MAP.put(ApproximationTyp.BSPLINE, aspBSpline);
        ASPECTPID_MAP.put(ApproximationTyp.CUBICSPLINE, aspCubicSpline);
        ASPECTPID_MAP.put(ApproximationTyp.POLYLINE, aspPolyline);
    }

    /**
     * Erfragt zu einem Approximationsverfahren die PID des Aspekts, unter dem Daten
     * gelernt werden sollen.
     *
     * @param verfahren das Approximationsverfahren.<br>
     *                  z.B.: {@link ApproximationTyp#BSPLINE},
     *                  {@link ApproximationTyp#CUBICSPLINE},
     *                  {@link ApproximationTyp#POLYLINE},
     *                  {@link ApproximationTyp#UNBESTIMMT}
     * @return die PID des Aspekts, unter dem Daten gelernt werden sollen.
     */
    public static String getAspect(final ApproximationTyp verfahren) {
        String pid = ASPECTPID_MAP.get(verfahren);
        if (pid == null) {
            return ASP_AGREGATION5_MINUTEN;
        }
        return pid;
    }
}
