/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.text.MessageFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.Semaphore;

import de.bsvrz.iav.gllernen.gllernen.domain.GanglinienLernenParameterLeser;
import de.bsvrz.iav.gllernen.gllernen.domain.GanglinienModellAutomatischesLernen;
import de.bsvrz.iav.gllib.gllib.domain.Archiv;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.shared.CronScheduler;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Verwaltet das Lernen und führt es zyklisch aus. Ist verantworlich für das
 * Überwachen der Lernparameter.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class Verwaltung {

    /** Die Menge der Messquerschnitte für die gelernt wird. */
    private final Set<MessQuerschnitt> messquerschnitte = new HashSet<>();

    /** Der Logger der Klasse. */
    private final Debug log = Debug.getLogger();

    /** Der Scheduler der das Lernen zyklisch anstößt. */
    private final CronScheduler scheduler;

    /** Die ID des Cron-Jobs für das Lernen. */
    private long jobId = -1;

    /** Der Zeitpunkt an dem das letzte Mal gelernt wurde. */
    private LocalDateTime letzteLernzeit;

    /** Verwaltet die maximale Anzahl gleichzeitiger Lernthreads. */
    private final Semaphore semaphore;

    private GlLernen applikation;

    private int slots;

    private Archiv archiv;

    /**
     * Initialisiert die Verwaltung. Die letzte Lernzeit wird mit heutiger Tag minus
     * Datenmindestalter initialisiert und der Scheduler für das Lernen gestartet.
     * 
     * @param ganglinienDb
     */
    Verwaltung(GlLernen applikation, GanglinienLernenParameterLeser parameterLeser, Archiv archiv,
            Collection<MessQuerschnitt> messQuerschnitte) {

        this.applikation = applikation;
        this.archiv = archiv;
        this.messquerschnitte.addAll(messQuerschnitte);

        slots = 2;

        // Anzahl der maximal gleichzeitig aktiven Lernthreads festlegen
        if (applikation.getMaxThreads() > 0) {
            // Anzahl durch Kommandozeilenargument festgelegt
            slots = applikation.getMaxThreads();
        } else if (System.getenv("NUMBER_OF_PROCESSORS") != null) {
            // Standard: doppelte Anzahl der vorhandenen (C)PUs
            slots = Integer.parseInt(System.getenv("NUMBER_OF_PROCESSORS")) * 2;
        } else {
            // Keine systemabhängige Festlegung möglich
            slots = 2;
        }
        semaphore = new Semaphore(slots);

        if (parameterLeser == null) {
            throw new IllegalStateException("Parameter zum Lernen von Ganglinien können nicht gelesen werden.");
        }

        scheduler = new CronScheduler();

        GanglinienModellAutomatischesLernen parameter = parameterLeser.getParameter();
        parameterAktualisiert(parameter);
        letzteLernzeit = LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT).minus(parameter.getDatenMindestalter())
                .minusDays(1);

        parameterLeser.parameterGeaendert().addHandler(p -> parameterAktualisiert(p));
        scheduler.start();

        log.config(() -> "Verwaltung bereit.");
    }

    private void parameterAktualisiert(GanglinienModellAutomatischesLernen parameter) {

        log.info(() -> MessageFormat.format("Neuen Parameter empfangen: {0}", parameter));
        if (jobId < 0) {
            // Parameter wird zum ersten Mal ausgewertet, ergo Job anlegen.
            jobId = scheduler.schedule(new LernTask(applikation, this, parameter));
        }

        LernTask lernTask = (LernTask) scheduler.getCronJob(jobId);
        lernTask.setLernParam(parameter);
    }

    /**
     * Gibt die Menge der Messquerschnitte zurück, für die gelernt wird.
     *
     * @return die Messquerschnitte.
     */
    public Set<MessQuerschnitt> getMessquerschnitte() {
        return messquerschnitte;
    }

    /**
     * Gibt den Zeitpunkt an dem das letzte Mal gelernt wurde zurück.
     *
     * @return der letzte Lernzeitpunkt.
     */
    public LocalDateTime getLetzteLernzeit() {
        return letzteLernzeit;
    }

    /**
     * Legt den Zeitpunkt fest, an dem das letzte Mal gelernt wurde.
     *
     * @param letzteLernzeit der letzte Lernzeitpunkt.
     */
    public void setLetzteLernzeit(final LocalDateTime letzteLernzeit) {
        log.fine(() -> MessageFormat.format("Neue letzte Lernzeit: {0}", letzteLernzeit));
        this.letzteLernzeit = letzteLernzeit;
    }

    /**
     * Gibt die Semaphore für Verwaltung der Lernthreads zurück.
     *
     * @return die Semaphore.
     */
    public Semaphore getSemaphore() {
        return semaphore;
    }

    public int getSlots() {
        return slots;
    }

    public Archiv getArchiv() {
        return archiv;
    }
}
